/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */


module clocks
#( 
   parameter  CLK200FX_MULT = 5,
   parameter  CLK200FX_DIV  = 8
)
(
   input      CLK200_N,
   input      CLK200_P,
   input      SYS_RESET_IN,

   output     CLK66,
   output     CLK200,
   output     CLK400,
   output     CLKFX,
   output reg RST
);

   

   wire       w_clk66_bufg_in,  w_clk66_bufg_out;
   wire       w_clk200_bufg_in, w_clk200_bufg_out;
   wire       w_clk400_bufg_in, w_clk400_bufg_out;
   wire       w_dcm_lock;
   

   //ibuf the reset and complement
   wire   w_sys_reset_in;
   wire   w_sys_reset;
   IBUF ibuf0( .I( SYS_RESET_IN), .O(w_sys_reset_in) );
   assign w_sys_reset = ~w_sys_reset_in;


   //ibuf the 200MHz clock
   wire   w_refclk200_in;
   IBUFGDS_LVPECL_25 lvpecl_clk200_in (
                                       .O   (w_refclk200_in),
                                       .I   (CLK200_P),
                                       .IB  (CLK200_N)
                                       );



   //Use a DCM to ramp all the signals down
   defparam DCM_BASEA.DLL_FREQUENCY_MODE = "HIGH";
   defparam DCM_BASEA.DUTY_CYCLE_CORRECTION = "TRUE";
   defparam DCM_BASEA.CLKDV_DIVIDE = 3.0; //200/3 = 66.7
   defparam DCM_BASEA.CLKFX_MULTIPLY = CLK200FX_MULT;//5;
   defparam DCM_BASEA.CLKFX_DIVIDE   = CLK200FX_DIV; //8;
   defparam DCM_BASEA.FACTORY_JF = 16'hF0F0;

   DCM_BASE DCM_BASEA (
                       .CLK0      (w_clk200_bufg_in),  // 200MHz
                       .CLK180    (),
                       .CLK270    (),
                       .CLK2X     (w_clk400_bufg_in),   // 2x 200MHz
                       .CLK2X180  (),
                       .CLK90     (),
                       .CLKDV     (w_clk66_bufg_in),
                       .CLKFX     (w_clkfx_bufg_in),
                       .CLKFX180  (),
                       .LOCKED    (w_dcm_lock),
                       .CLKFB     (w_clk200_bufg_out),
                       .CLKIN     (w_refclk200_in),
                       .RST       (w_sys_reset)
                       );


   BUFG dcm_clk66(
                   .I (w_clk66_bufg_in),
                   .O (w_clk66_bufg_out)
                   );

   
   //This one used for feedback
   BUFG dcm_clk200(
                   .I (w_clk200_bufg_in),
                   .O (w_clk200_bufg_out)
                   );


   BUFG dcm_clk400(
                   .I(w_clk400_bufg_in),
                   .O(w_clk400_bufg_out)
                   );

   BUFG dcm_clkfx(
                   .I (w_clkfx_bufg_in),
                   .O (w_clkfx_bufg_out)
                   );



   //Base the reset off of the dcm lock
   reg    sys_rst_0, sys_rst_1, sys_rst_2;
   
   always@(posedge w_clk200_bufg_out) begin
      if(w_dcm_lock == 1'b0) begin
         sys_rst_0 <= 1'b1;
         sys_rst_1 <= 1'b1;
         sys_rst_2 <= 1'b1;
         RST       <= 1'b1;
      end else begin
         sys_rst_0 <= 1'b0;
         sys_rst_1 <= sys_rst_0;
         sys_rst_2 <= sys_rst_1;
         RST       <= sys_rst_2;
      end 
   end

   assign CLK66  = w_clk66_bufg_out;
   assign CLK200 = w_clk200_bufg_out;
   assign CLK400 = w_clk400_bufg_out;
   assign CLKFX  = w_clkfx_bufg_out;
   
   

   
endmodule
   
