/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: LLNL
// Engineer: Weirong Jiang
// 
// Create Date:    12:47:14 02/23/2010 
// Design Name: 
// Module Name:    lshc 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: LSH-based classifier
//						When two packets are coming, each is fed into one port of LSH-C
//						When only one packet is coming, that packet is fed into both ports
//						When no packet is coming, valid is 0
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module lshc(
    enable,
	 ipproto,
	 pktsize,
    srcport,
    dstport,
	 ipprotoB,
    pktsizeB,
    srcportB,
    dstportB,
    clk,
    appidi,
    we,
    appido,
	 appidoB
    );

function integer log2;
  input integer value;
  begin
    value = value-1;
    for (log2=0; value>0; log2=log2+1)
      value = value>>1;
  end
endfunction

parameter N_HASHTB = 16; // # of hash tables
parameter TB_ADDR_W = 10; // address width of each table
parameter IPPROTO_W = 8; // # of bits in ip.protocol
parameter FEATURE_W = 16; // # of bits per feature
parameter APPID_W = 2; // # of applications to be classified: 2^APPID_W 
parameter H_MASKS = {
	56'h0000c000f000f0,
	56'h000c000f000f00,
	56'h000000f800f800,
	56'h0000007c007c00,

	56'h00000007e0002e, 
	56'h000000002e07e0,
	56'h0000000e70003c,
	56'h000000003c0e70,
	56'h0000f0000e000e,
	56'h0000300f000f00,
	56'h0000c00f000f00,
	56'h000000003e003e,
	56'h0000280ce00016,
	56'h00002800160ce0,
	56'h0000600f00cc00,
	56'h000060cc000f00	
	};
parameter H_TYPES = 16'b1111000000000000; // 1 - UDP, 0 - TCP
parameter K = 1; // K-NN
parameter DISTANCE_W = log2(FEATURE_W)+2; // Sum of 1 distance with log2(IPPROTO_W) bits plus 3 distances each of which has log2(FEATURE_W)+1 bits. The maximum distance sum is 8+16*3=56, which needs 6 bits to represent.
parameter MASK_W = IPPROTO_W+FEATURE_W*3;
parameter N_SORT = (1<<log2(N_HASHTB));

input [IPPROTO_W-1:0] ipproto, ipprotoB;
input [FEATURE_W-1:0] pktsize, srcport, dstport;
input [FEATURE_W-1:0] pktsizeB, srcportB, dstportB;
input enable, clk, we;
input [APPID_W-1:0] appidi;
output [APPID_W-1:0] appido;
output [APPID_W-1:0] appidoB;

wire [APPID_W-1:0] appo [N_HASHTB-1:0]; // outputs from hash tables
wire [DISTANCE_W-1:0] dsto [N_HASHTB-1:0]; 
wire [APPID_W*N_SORT-1:0] appo_unsorted; // inputs to sorting
wire [DISTANCE_W*N_SORT-1:0] dsto_unsorted;
wire [APPID_W*N_SORT-1:0] appo_sorted; // descending order: the bottom has the shortest distance

wire [APPID_W-1:0] appoB [N_HASHTB-1:0];
wire [DISTANCE_W-1:0] dstoB [N_HASHTB-1:0];
wire [APPID_W*N_SORT-1:0] appo_unsortedB;
wire [DISTANCE_W*N_SORT-1:0] dsto_unsortedB;
wire [APPID_W*N_SORT-1:0] appo_sortedB; // descending order: the bottom has the shortest distance

assign appido = majority(appo_sorted[APPID_W*N_HASHTB-1:0], K);
assign appidoB = majority(appo_sortedB[APPID_W*N_HASHTB-1:0], K);

// hash tables
generate
	genvar i;
	for(i = 0; i < N_HASHTB; i = i+1) begin : gen_hashtb
		lshtb #(
		.HASH_MASK(H_MASKS[i*MASK_W +: MASK_W]),
		.HASH_TYPE(H_TYPES[i]),
		.TB_ADDR_W(TB_ADDR_W),
		.IPPROTO_W(IPPROTO_W),
		.FEATURE_W(FEATURE_W),
		.APPID_W(APPID_W)
		) htb(
		.enable(enable),
		.ipproto(ipproto),
		.pktsizei(pktsize), 
		.srcporti(srcport), 
		.dstporti(dstport),
		.ipprotoB(ipprotoB),
		.pktsizeiB(pktsizeB), 
		.srcportiB(srcportB), 
		.dstportiB(dstportB),
		.clk(clk),
		.appidi(appidi),
		.we(we),
		.appido(appo[i]),
		.distanceo(dsto[i]),
		.appidoB(appoB[i]),
		.distanceoB(dstoB[i])
		);

		assign appo_unsorted[i*APPID_W +: APPID_W] = appo[i];
		assign dsto_unsorted[i*DISTANCE_W +: DISTANCE_W] = dsto[i];
	
		assign appo_unsortedB[i*APPID_W +: APPID_W] = appoB[i];
		assign dsto_unsortedB[i*DISTANCE_W +: DISTANCE_W] = dstoB[i];
	end
	
	// in case the # of hash tables is not some power of 2
	for(i = N_HASHTB; i < N_SORT; i = i+1) begin : gen_sortmore
		assign appo_unsorted[i*APPID_W +: APPID_W] = 0;
		assign dsto_unsorted[i*DISTANCE_W +: DISTANCE_W] = -1; 
		assign appo_unsortedB[i*APPID_W +: APPID_W] = 0;
		assign dsto_unsortedB[i*DISTANCE_W +: DISTANCE_W] = -1; 
	end
endgenerate

// sorting modules
sorting #(
	.ID_W(APPID_W),
	.VALUE_W(DISTANCE_W),
	.N_ENTRY(N_SORT)
	) sm(
	.idi(appo_unsorted),
	.vi(dsto_unsorted),
	.clk(clk),
	.ido(appo_sorted)
	);

sorting #(
	.ID_W(APPID_W),
	.VALUE_W(DISTANCE_W),
	.N_ENTRY(N_SORT)
	) smB(
	.idi(appo_unsortedB),
	.vi(dsto_unsortedB),
	.clk(clk),
	.ido(appo_sortedB)
	);

// majority voter
function [APPID_W-1:0] majority;
	input [APPID_W*N_HASHTB-1:0] appi;
	input integer K;
	integer i;
	reg [APPID_W-1:0] apps [N_HASHTB-1:0];
	reg [log2(N_HASHTB)-1:0] nv [(1<<APPID_W)-1:0];
	reg [APPID_W-1:0] mapp;
	begin
		mapp = 0;
		for(i = 0; i < (1<<APPID_W) ; i = i+1) begin
			nv[i] = 0;
		end
		for(i = 0; (i < K) && (i < N_HASHTB); i = i+1) begin
			apps[i] = appi[i*APPID_W +: APPID_W];
			nv[apps[i]] = nv[apps[i]] + 1;
			if(nv[apps[i]] > nv[mapp]) begin
				mapp = apps[i];
			end
		end
		majority = mapp;
	end
endfunction

endmodule
