/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: LLNL
// Engineer: Weirong Jiang 
// 
// Create Date:    13:14:15 02/23/2010 
// Design Name: 
// Module Name:    lshtb 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: LSH table
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module lshtb(
    enable,
	 ipproto,
	 pktsizei,
    srcporti,
    dstporti,
	 ipprotoB,
    pktsizeiB,
    srcportiB,
    dstportiB,
    clk,
    appidi,
    we,
    appido,
	 distanceo,
    appidoB,
	 distanceoB
    );

function integer log2;
  input integer value;
  begin
    value = value-1;
    for (log2=0; value>0; log2=log2+1)
      value = value>>1;
  end
endfunction

parameter HASH_MASK = 'h00000F00070007; // mask to generate hash key
parameter HASH_TYPE = 0; // hash type = 0, TCP; 1, UDP.
parameter TB_ADDR_W = 10; // address width of the table
parameter IPPROTO_W = 8; // # of bits in ip.protocol
parameter FEATURE_W = 16; // # of bits per feature
parameter APPID_W = 2; // # of applications to be classified: 2^APPID_W 

parameter TB_DEPTH = 1<<TB_ADDR_W; // depth of the table
parameter TB_WIDTH = 1 + FEATURE_W * 3 + APPID_W; // width of the table
parameter DISTANCE_W = log2(FEATURE_W) + 2; //Hamming: log2(FEATURE_W)+1; 
// Sum of 3 items each of which has log2(FEATURE_W)+1 bits
parameter MASK_W = IPPROTO_W+FEATURE_W*3;

input [IPPROTO_W-1:0] ipproto, ipprotoB;
input [FEATURE_W-1:0] pktsizei, srcporti, dstporti;
input [FEATURE_W-1:0] pktsizeiB, srcportiB, dstportiB;
input enable, clk, we;
input [APPID_W-1:0] appidi;

output [APPID_W-1:0] appido;
output [DISTANCE_W-1:0] distanceo; 
output [APPID_W-1:0] appidoB;
output [DISTANCE_W-1:0] distanceoB; 

wire ipptypeo;
wire [FEATURE_W-1:0] pktsizeo, srcporto, dstporto;
wire [APPID_W-1:0] appidr;
reg ipptyper;
reg [FEATURE_W-1:0] pktsizer, srcportr, dstportr;
reg [APPID_W-1:0] appido;
reg [DISTANCE_W-1:0] distanceo; 

wire ipptypeoB;
wire [FEATURE_W-1:0] pktsizeoB, srcportoB, dstportoB;
wire [APPID_W-1:0] appidrB;
reg ipptyperB;
reg [FEATURE_W-1:0] pktsizerB, srcportrB, dstportrB;
reg [APPID_W-1:0] appidoB;
reg [DISTANCE_W-1:0] distanceoB; 

reg [TB_WIDTH-1:0] ram [TB_DEPTH-1:0];

wire ipptype, ipptypeB;
wire [TB_ADDR_W-1:0] haddr;
wire [TB_ADDR_W-1:0] haddrB;
reg [TB_ADDR_W-1:0] rhaddr;
reg [TB_ADDR_W-1:0] rhaddrB;

assign ipptype = (ipproto == 8'h11);
assign ipptypeB = (ipprotoB == 8'h11);

assign haddr = (ipptype == HASH_TYPE) ? hkey(ipproto, pktsizei, srcporti, dstporti, HASH_MASK) : hkey(ipproto, pktsizei, 0, 0, HASH_MASK);
assign haddrB = (ipptypeB == HASH_TYPE) ? hkey(ipprotoB, pktsizeiB, srcportiB, dstportiB, HASH_MASK) : hkey(ipprotoB, pktsizeiB, 0, 0, HASH_MASK);

// Signal relay & Output
always @(posedge clk)
begin
	{ipptyper, pktsizer, srcportr, dstportr} <= {ipptype, pktsizei, srcporti, dstporti};
	distanceo <= (ipptyper == ipptypeo) ? (hammd(srcporto, srcportr) + hammd(dstporto, dstportr) + hammd(pktsizeo, pktsizer)) : -1;
	appido <= appidr;
	{ipptyperB, pktsizerB, srcportrB, dstportrB} <= {ipptypeB, pktsizeiB, srcportiB, dstportiB};
	distanceoB <= (ipptyperB == ipptypeoB) ? (hammd(srcportoB, srcportrB) + hammd(dstportoB, dstportrB) + hammd(pktsizeoB, pktsizerB)) : -1;
	appidoB <= appidrB;
end

// RAM
always @(posedge clk)
begin
	if(enable) begin
		if(we) begin 
			ram[haddr] <= {ipptype, pktsizei, srcporti, dstporti, appidi};
		end
		rhaddr <= haddr;
		rhaddrB <= haddrB;
	end
end
assign {ipptypeo, pktsizeo, srcporto, dstporto, appidr} = ram[rhaddr];
assign {ipptypeoB, pktsizeoB, srcportoB, dstportoB, appidrB} = ram[rhaddrB];


// Hash key generation
function [TB_ADDR_W-1:0] hkey;
	 input [IPPROTO_W-1:0] ipproto;
	 input [FEATURE_W-1:0] pktsize, srcport, dstport;
	 input [MASK_W-1:0] hmask;
	 integer i, j;
	 reg [MASK_W-1:0] temp;
	 begin
		temp = {ipproto, pktsize, srcport, dstport};
		j = TB_ADDR_W-1;
		for(i = MASK_W-1; (i >= 0)&&(j >= 0); i = i-1) begin
			if(hmask[i]) begin
				hkey[j] = temp[i];
				j = j-1;
			end
		end
	 end
endfunction

// Compute the hamming distance
function [log2(FEATURE_W):0] hammd;
	input [FEATURE_W-1:0] dia, dib;
	reg [FEATURE_W-1:0] axorb;
	reg [log2(FEATURE_W):0] temp;
	integer i;
	begin
		axorb = dia ^ dib;
		temp = 0;
		for(i = 0; i < FEATURE_W; i = i+1) begin
			temp = temp + axorb[i];
		end
		hammd = temp;
	end
endfunction


endmodule









